# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019, 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

"""Unsupervised nearest neighbors learner"""

from pai4sk.neighbors._unsupervised import NearestNeighborsSklearn
import warnings
import numpy as np
from .utils import _can_allow_cuml, _is_cuml_algorithm, _do_not_allow_cuml


class NearestNeighbors(NearestNeighborsSklearn):
    """Unsupervised learner for implementing neighbor searches.

    If cuml is installed and input data is cudf dataframe, then pai4sk will try
    to use the accelerated KNN algorithm from cuML. Otherwise, scikit-learn's
    KNN algorithm will be used.

    cuML in pai4sk is currently supported only without MPI.
    |  If KNN from cuML is run, then the return values from the APIs will be
    cudf dataframe and cudf Series objects instead of the return types of
    scikit-learn API.

    Read more in the :ref:`User Guide <unsupervised_neighbors>`.

    Parameters
    ----------
    n_neighbors : int, optional (default = 5)
        Number of neighbors to use by default for :meth:`kneighbors` queries.

    radius : float, optional (default = 1.0)
        Range of parameter space to use by default for :meth:`radius_neighbors`
        queries.

    algorithm : {'auto', 'ball_tree', 'kd_tree', 'brute', 'cuml'}, optional, default is 'auto'.
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`BallTree`
        - 'kd_tree' will use :class:`KDTree`
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm
          based on the values passed to :meth:`fit` method.

        Note: fitting on sparse input will override the setting of
        this parameter, using brute force.

        auto or brute: If cuml is installed and input data is cudf dataframe, then pai4sk
              will try to use KNN algorithm from RAPIDS cuML if possible. Otherwise
              NearestNeighbors from scikit-learn will be used.
        cuml: If cuml is installed and input data is cudf dataframe, then pai4sk
              will try to use KNN algorithm from RAPIDS cuML if possible. Otherwise
              throws an error.

        cuML in pai4sk is currently supported only without MPI.

    use_gpu : boolean, Default is True
        If True, cuML will use all GPUs. Applicable only for cuML.

    handle : cuml.Handle, Default is None
        The cumlHandle resources to use.
        If it is None, a new one is created just for this class. Applicable only for cuML.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.

    metric : string or callable, default 'minkowski'
        metric to use for distance computation. Any metric from scikit-learn
        or scipy.spatial.distance can be used.

        If metric is a callable function, it is called on each
        pair of instances (rows) and the resulting value recorded. The callable
        should take two arrays as input and return one value indicating the
        distance between them. This works for Scipy's metrics, but is less
        efficient than passing the metric name as a string.

        Distance matrices are not supported.

        Valid values for metric are:

        - from scikit-learn: ['cityblock', 'cosine', 'euclidean', 'l1', 'l2',
          'manhattan']

        - from scipy.spatial.distance: ['braycurtis', 'canberra', 'chebyshev',
          'correlation', 'dice', 'hamming', 'jaccard', 'kulsinski',
          'mahalanobis', 'minkowski', 'rogerstanimoto', 'russellrao',
          'seuclidean', 'sokalmichener', 'sokalsneath', 'sqeuclidean',
          'yule']

        See the documentation for scipy.spatial.distance for details on these
        metrics.

    p : integer, optional (default = 2)
        Parameter for the Minkowski metric from
        pai4sk.metrics.pairwise.pairwise_distances. When p = 1, this is
        equivalent to using manhattan_distance (l1), and euclidean_distance
        (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.

    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.

    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    should_downcast: Bool (default=True)
        Currently only single precision is supported in the underlying undex. Setting this to
        true will allow single-precision input arrays to be automatically downcasted to single
        precision. Default = True. Applicable for cuML only.

    verbose: Bool (default=False)
        Verbose output. Applicable for cuML only.

    Examples
    --------
      >>> import numpy as np
      >>> from pai4sk.neighbors import NearestNeighbors
      >>> samples = [[0, 0, 2], [1, 0, 0], [0, 0, 1]]

      >>> neigh = NearestNeighbors(2, 0.4)
      >>> neigh.fit(samples)  #doctest: +ELLIPSIS
      NearestNeighbors(...)

      >>> neigh.kneighbors([[0, 0, 1.3]], 2, return_distance=False)
      ... #doctest: +ELLIPSIS
      array([[2, 0]]...)

      >>> nbrs = neigh.radius_neighbors([[0, 0, 1.3]], 0.4, return_distance=False)
      >>> np.asarray(nbrs[0][0])
      array(2)

    See also
    --------
    KNeighborsClassifier
    RadiusNeighborsClassifier
    KNeighborsRegressor
    RadiusNeighborsRegressor
    BallTree

    Notes
    -----
    See :ref:`Nearest Neighbors <neighbors>` in the online documentation
    for a discussion of the choice of ``algorithm`` and ``leaf_size``.

    https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm
    """

    __slots__ = ('_NearestNeighborsSklearn')

    def __init__(self, n_neighbors = 5, use_gpu = True, verbose = False,
                 should_downcast = True, radius=1.0, algorithm='auto', leaf_size=30,
                 metric='minkowski', p=2, metric_params=None, n_jobs=None, handle=None, **kwargs):
        self.use_gpu = use_gpu
        self.n_neighbors = n_neighbors
        self.algorithm = algorithm
        self.verbose = verbose
        self.metric = metric
        self.handle = handle
        self.should_downcast = should_downcast
        self._CUML_ARGS = ['use_gpu', 'verbose', 'should_downcast', 'handle']
        self._NearestNeighborsSklearn=super(NearestNeighbors, self)
        self._cumlknn = None
        self._NearestNeighborsSklearn.__init__(n_neighbors=n_neighbors,
                                               radius=radius,
                                               algorithm=algorithm,
                                               leaf_size=leaf_size, metric=metric,
                                               p=p, metric_params=metric_params,
                                               n_jobs=n_jobs, **kwargs)
        # Perform basic validations for enabling cuML first
        self._cuml_basic_validations = _can_allow_cuml()

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._NearestNeighborsSklearn = super(NearestNeighbors, self)

    def fit(self, X, y=None):
        """Fit the model using X as training data

        Parameters
        ----------
        X : {array-like, sparse matrix, BallTree, KDTree} or cudf dataframe
            Training data. If array or matrix, shape [n_samples, n_features],
            or [n_samples, n_samples] if metric='precomputed'.
        """

        if self._check_set_cuml(X):
            print("Info: Invoking cuML KNN fit...")
            self._fit_X = X
            self._cumlknn.fit(X)
        else:
            self._NearestNeighborsSklearn.fit(X, y)

        return self

    def kneighbors(self, X=None, n_neighbors=None, return_distance=True):
        """Finds the K-neighbors of a point.
        Returns indices of and distances to the neighbors of each point.

        Parameters
        ----------
        X : cudf dataframe or array-like, shape (n_query, n_features), \
                or (n_query, n_indexed) if metric == 'precomputed'
            The query point or points.
            If not provided, neighbors of each indexed point are returned.
            In this case, the query point is not considered its own neighbor.

        n_neighbors : int
            Number of neighbors to get (default is the value
            passed to the constructor).

        return_distance : boolean, optional. Defaults to True.
            If False, distances will not be returned

        Returns
        -------
        dist : array or cudf dataframe
            Array representing the lengths to points, only present if
            return_distance=True

        ind : array or cudf dataframe
            Indices of the nearest points in the population matrix.

        If KMeans from cuML is run then this method returns cudf dataframes
        instead of the return types of scikit-learn API.
        """

        if (n_neighbors is None):
            n_neighbors = self.n_neighbors

        if n_neighbors <= 0:
            raise ValueError("Expected n_neighbors > 0. Got %d" % n_neighbors)

        if not np.issubdtype(type(n_neighbors), np.integer):
            raise TypeError("n_neighbors does not take %s value, "
                            "enter integer value" % type(n_neighbors))

        if self._cumlknn:
            if X is None:
                X = self._fit_X

            train_size = X.shape[0]
            if n_neighbors > train_size:
                raise ValueError("Expected n_neighbors <= n_samples, "
                    " but n_samples = %d, n_neighbors = %d" %
                    (train_size, n_neighbors))

            D, I = self._cumlknn.kneighbors(X, n_neighbors)
            if return_distance:
                return (D, I)
            else:
                return I
        else:
            _do_not_allow_cuml(X)
            return self._NearestNeighborsSklearn.kneighbors(X, n_neighbors, return_distance)

    # Override get_params of BaseEstimator
    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        params_dict = super(NearestNeighbors, self).get_params(deep)
        # Remove cuml-specific new args from the dict
        for cuml_arg in self._CUML_ARGS:
            params_dict.pop(cuml_arg, None)

        return params_dict

    def _check_set_cuml(self, X):
        """
           Check if cuml can be used for this NearestNeighbors
        """

        # Check if basic validations for enabling cuML succeeded
        if not self._cuml_basic_validations:
            return False

        # Perform further validations for enabling cuML
        import cudf
        if not isinstance(X, cudf.DataFrame):
            warnings.warn("Currently cuML in pai4sk can work only on " \
                          "cudf.DataFrame.", UserWarning)
            return False

        if not self.use_gpu:
            warnings.warn("Conflicting arguments are seen. cudf.DataFrame" \
                          " is given as input and use_gpu is set to False.",
                          UserWarning)
            return False

        if not _is_cuml_algorithm(self.algorithm, ['brute']):
            return False

        # cuml's KNN currently only supports euclidean as metric and that is the default value
        if self.metric is 'minkowski':
            self.metric = 'euclidean'
        elif self.metric is not 'euclidean':
            warnings.warn("Currently cuML KNN supports only euclidean metric." \
                          " Conflicting arguments are seen. cudf.DataFrame is" \
                          " given as input and metric is not euclidean.", UserWarning)
            return False

        if not self._cumlknn:
            from pai4sk import sml_utils
            ndevices = sml_utils.get_gpu_count()
            if ndevices == 0:
                warnings.warn("Currently cuML in pai4sk can work only with GPUs." \
                              "No GPUs are available.", UserWarning)
                return False

            from cuml import NearestNeighbors as cumlknn
            self._cumlknn = cumlknn(handle=self.handle, n_neighbors=self.n_neighbors,
                                    verbose=self.verbose, algorithm=self.algorithm,
                                    metric=self.metric)
            warnings.warn("Using KNN from cuML.", UserWarning)

        return True

