# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2018, 2020. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

import time
import sys
import os
import numpy as np
import argparse
from sklearn.metrics import roc_auc_score

sys.path.insert(0, os.path.abspath('../'))

CLI=argparse.ArgumentParser()

CLI.add_argument(
    "--skip_xgboost",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--data_path",
    type=str,
    default="."
)
CLI.add_argument(
    "--num_threads",
    type=int,
    default=16
)
CLI.add_argument(
    "--use_gpu",
    action = 'store_true',
    default=False
)

args = CLI.parse_args()

# Load the input datasets
t0 = time.time()
X_train = np.load(args.data_path + "/data/epsilon.X_train.npy")
X_test  = np.load(args.data_path + "/data/epsilon.X_test.npy")
y_train = np.load(args.data_path + "/data/epsilon.y_train.npy")
y_test  = np.load(args.data_path + "/data/epsilon.y_test.npy")
print("Data load time (s):  {0:.2f}".format(time.time()-t0))

# Convert the -1,1 labels to 0,1 required by xgboost (v0.9)
y_test  = (y_test + 1)/2
y_train = (y_train + 1)/2

# Boosting parameters
max_depth = 6
num_round = 100
learning_rate = 0.1

# Import BoostingMachine from snap_ml or pai4sk
from pai4sk import BoostingMachine

# In constrast to other boosting frameworks, pai4sk's BoostingMachine model dose
# not utilize a fixed maximal tree depth at each boosting iteration.
# Instead, the tree depth is sampled at each boosting iteration according to a
# discrete uniform distribution from [min_max_depth, max_max_depth].
# For a fair runtime evaluation against xgboost we set max_max_depth to be equal to min_max_depth

# Create a pai4sk BoostingMachine object
booster = BoostingMachine(objective='logloss', num_round=num_round, min_max_depth=max_depth, max_max_depth=max_depth,
                             learning_rate=learning_rate, random_state=42, use_gpu=args.use_gpu, n_threads=args.num_threads)

# Train a pai4sk BoostingMachine model
t0 = time.time()
booster.fit(X_train, y_train)
t_fit = time.time()-t0

# Predict with the BoostingMachine model
yhat_test = booster.predict(X_test)

if args.use_gpu == False:
    print("[pai4sk - BoostingMachine-CPU] training_time %f roc_auc_score %f " % (t_fit,roc_auc_score(y_test, yhat_test)))
else:
    print("[pai4sk - BoostingMachine-GPU] training_time %f roc_auc_score %f " % (t_fit,roc_auc_score(y_test, yhat_test)))


if args.skip_xgboost:
    print('Execution using xgboost is skipped.')
    exit()


# Import xgbooster from XGBoost
from xgboost import XGBClassifier

booster = XGBClassifier(n_estimators=num_round, max_depth=max_depth, learning_rate=learning_rate,
                            random_state=42, tree_method = 'gpu_hist' if args.use_gpu else 'hist', n_jobs=args.num_threads,
                            reg_lambda=0, min_child_weight=0)

# Train XGB
t0 = time.time()
booster.fit(X_train, y_train)
t_fit = time.time()-t0

# Predict XGB
yhat_test = booster.predict(X_test, output_margin=True)

if args.use_gpu == False:
    print("[XGBoost-CPU] training_time %f roc_auc_score %f " % (t_fit,roc_auc_score(y_test, yhat_test)))
else:
    print("[XGBoost-GPU] training_time %f roc_auc_score %f " % (t_fit,roc_auc_score(y_test, yhat_test)))

