# Epsilon Example

In this example we will train a Support Vector Machine model on the Epsilon dataset, using both `scikit-learn` (`sklearn`) and `pai4sk`. Note that `scikit-learn` does not provide an MPI implementation. We will however still perform this comparison in order to mainly verify the accuracy metrics. 

The epsilon dataset is from the [PASCAL Large Scale Learning Challenge](http://www.k4all.org/project/large-scale-learning-challenge/). 

The challenge is described as follows:

> This PASCAL Challenge is concerned with the scalability and efficiency of existing ML approaches with respect to computational, memory or communication resources, e.g. resulting from a high algorithmic complexity, from the size or dimensionality of the data set, and from the trade-off between distributed resolution and communication costs.

## Download the Data

Download and decompress the data from the LIBSVM repository:
```bash
mkdir data
cd data
wget https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary/epsilon_normalized.bz2
bunzip2 epsilon_normalized.bz2
cd ../
```
Now one should see a file named `epsilon_normalized` in the `data/` directory.

## Data Pre-processing
The data is in SvmLight format which is not really appropriate since this dataset is dense. Therefore, we perform the following pre-processing steps. 

Firstly, we load the data from SvmLight format. Next we split the data into train and test datasets.
Then we convert the newly created datasets to dense numpy arrays since the dataset is dense.

Next we write the dense data structures to disk as binary numpy files. This helps to achieve fast data loading.
This format will be used for loading the datasets using `sklearn`. `sklearn` will be used for performance comparison.

Finally, we write the train and test datasets in the distributed snap format which will be used only for training the `pai4sk` Support Vector Machine model.
This format helps to achieve higher performance.

A summary of all the pre-processed files is shown below: 
```bash
ls -l data/
epsilon_normalized
epsilon.train.snap
epsilon.test.snap
epsilon.X_train.npy
epsilon.X_test.npy
epsilon.y_train.npy
epsilon.y_train.npy
```

The entire pre-processing code can be run by executing `snaprun $CONDA_PREFIX/pai4sk/mpi-examples/example-epsilon/preprocess-epsilon.py`


## Training and Evaluating a Support Vector Machine Model

After running the data pre-processing steps above, we are ready to perform our benchmarking: `pai4sk` vs. `sklearn`.

First, we train a Support Vector Machine model using the IBM `pai4sk` library. We provide command line option --use_gpu to run the example for how to train the model using 1) GPUs or 2) CPUs. 

We start by defining the data paths for the train and test datasets in snap format.

Also you need to have access to the data from all the connected machines that will be used (by replicating the data or by providing a distributed file system).

Next, we read the datasets from snap format using the loading functions provided by the `pai4sk` library and we measure the data loading time.

After initialization, we call the `fit()` function in order to train the dual Support Vector Machine model. We also time the execution of the training process.

Please note that in order to achieve maximum performance the `pai4sk` library assumes 1 MPI process per node.

As a last step we make predictions on the test data and we compute the accuracy metric. The IBM `pai4sk` library provides a distributed
implementation of the accuracy metrics.

Finally, we train a Support Vector Machine model using the `scikit-learn` (`sklearn`) library.

## Results

To test the code on 1 machine with 1 MPI process, one should run the following command: 
```bash
snaprun python $CONDA_PREFIX/pai4sk/mpi-examples/example-epsilon/example-epsilon.py 
```

To test the code on 2 machines with 1 MPI process per machine, one should run the following command:
```bash
snaprun -H host1,host2 python $CONDA_PREFIX/pai4sk/mpi-examples/example-epsilon/example-epsilon.py
```
The user can also run multiple MPI processes per machine, but, in that case, the MPI processes on one machine will share the same hardware resources during training.
Namely, if the user trains the model on GPUs, the MPI processes on one machine will share the same GPUs as defined in the `device_ids` parameter. Therefore, we recommend
the user to run 1 MPI process per machine in order to achieve the maximum performance. 

One should see the following output on a **P9+V100** system when running `snaprun python $CONDA_PREFIX/pai4sk/mpi-examples/example-epsilon/example-epsilon.py`:
```
[pai4sk] [GPU Training] Accuracy = 0.898
[pai4sk] [GPU Training] Training time = 3.135

[sklearn] Accuracy = 0.898
[sklearn] Training time = 24.770
```

One should see the following results on 2 **P9+V100** systems when running `snaprun -H host1,host2 python $CONDA_PREFIX/pai4sk/mpi-examples/example-epsilon/example-epsilon.py:
```
[pai4sk] [GPU Training] Accuracy = 0.898
[pai4sk] [GPU Training] Training time = 2.837

[sklearn] Accuracy = 0.898
[sklearn] Training time = 24.770
```

The reported times can vary depending on the system architecture.

&copy; Copyright IBM Corporation 2018, 2019
